function heuristic(nodeA, nodeB)
	return math.sqrt(math.pow(nodeB.x - nodeA.x, 2) + math.pow(nodeB.y - nodeA.y, 2))
end

function best_f_score(set, f_score)
	local lowest = 1/0
  local bestNode = nil
	for _, node in ipairs(set) do
		local score = f_score[NodeToString(node)]
		if score < lowest then
			lowest = score
      bestNode = node
		end
	end
	return bestNode
end

function neighbor_nodes(node, map)
	local neighbors = {}  
  if map[node.x + 1][node.y] ~= '1' then
    table.insert(neighbors, {x = node.x + 1, y = node.y})
  end
  if map[node.x - 1][node.y] ~= '1' then
    table.insert(neighbors, {x = node.x - 1, y = node.y})
  end
  if map[node.x][node.y + 1] ~= '1' then
    table.insert(neighbors, {x = node.x, y = node.y + 1})
  end
  if map[node.x][node.y - 1] ~= '1' then
    table.insert(neighbors, {x = node.x, y = node.y - 1})
  end
  return neighbors
end

function not_in(set, node)
	for _, nd in ipairs (set) do
		if nd.x == node.x and nd.y == node.y then 
      return false 
    end
	end
	return true
end

function remove_node(set, node)
	for i, nd in ipairs ( set ) do
		if nd.x == node.x and nd.y == node.y then 
			set[i] = set[#set]
			set[#set] = nil
			break
		end
	end	
end

function reverse_path(r_path, map, current_node)
	if map[NodeToString(current_node)] then
		table.insert(r_path, 1, map[NodeToString(current_node)]) 
		return reverse_path(r_path, map, map[NodeToString(current_node)])
	else
		return r_path
	end
end

function NodeToString(node)
  return node.x .. " " .. node.y
end

function astar_path(start, goal, map)
	local closedset = {}
	local openset = {start}
	local came_from = {}
	local g_score = {}
  local f_score = {}
  
	g_score[NodeToString(start)] = 0
	f_score[NodeToString(start)] = g_score[NodeToString(start)] + heuristic(start, goal)

	while #openset > 0 do
    
		local current = best_f_score(openset, f_score)
		if current.x == goal.x and current.y == goal.y then
			local path = reverse_path ({}, came_from, goal)
			table.insert(path, goal)
			return path
		end

		remove_node(openset, current)		
		table.insert(closedset, current)
		
		local neighbors = neighbor_nodes(current, map)
		for _, neighbor in ipairs (neighbors) do 
			if not_in (closedset, neighbor) then
			
				local tentative_g_score = g_score [NodeToString(current)] + heuristic(current, neighbor)
				 
				if not_in(openset, neighbor) or tentative_g_score < g_score[NodeToString(neighbor)] then 
					came_from[NodeToString(neighbor)] = current
					g_score[NodeToString(neighbor)] = tentative_g_score
					f_score[NodeToString(neighbor)] = g_score[NodeToString(neighbor)] + heuristic(neighbor, goal)
					if not_in(openset, neighbor) then
						table.insert (openset, neighbor)
					end
				end
			end
		end
	end
	return nil 
end